@@ -12,7 +12,11 @@
 #undef _DARWIN_C_SOURCE
 #endif
 #include "data-types.h"
+#include "cleanup.h"
+#include "safe-wrappers.h"
 #include "control-codes.h"
+#include "wcwidth-std.h"
+#include "wcswidth.h"
 #include "modes.h"
 #include <stddef.h>
 #include <termios.h>
@@ -38,18 +42,17 @@ static PyObject*
 process_group_map() {
     int num_of_processes = proc_listallpids(NULL, 0);
     size_t bufsize = sizeof(pid_t) * (num_of_processes + 1024);
-    pid_t *buf = malloc(bufsize);
+    FREE_AFTER_FUNCTION pid_t *buf = malloc(bufsize);
     if (!buf) return PyErr_NoMemory();
     num_of_processes = proc_listallpids(buf, (int)bufsize);
     PyObject *ans = PyTuple_New(num_of_processes);
-    if (ans == NULL) { free(buf); return PyErr_NoMemory(); }
+    if (ans == NULL) { return PyErr_NoMemory(); }
     for (int i = 0; i < num_of_processes; i++) {
         long pid = buf[i], pgid = getpgid(buf[i]);
         PyObject *t = Py_BuildValue("ll", pid, pgid);
-        if (t == NULL) { free(buf); Py_DECREF(ans); return NULL; }
+        if (t == NULL) { Py_DECREF(ans); return NULL; }
         PyTuple_SET_ITEM(ans, i, t);
     }
-    free(buf);
     return ans;
 }
 #endif
@@ -89,7 +92,7 @@ stop_profiler(PyObject UNUSED *self, PyObject *args UNUSED) {
 #endif
 
 static inline bool
-put_tty_in_raw_mode(int fd, const struct termios* termios_p, bool read_with_timeout) {
+put_tty_in_raw_mode(int fd, const struct termios* termios_p, bool read_with_timeout, int optional_actions) {
     struct termios raw_termios = *termios_p;
     cfmakeraw(&raw_termios);
     if (read_with_timeout) {
@@ -97,42 +100,42 @@ put_tty_in_raw_mode(int fd, const struct termios* termios_p, bool read_with_time
     } else {
         raw_termios.c_cc[VMIN] = 1; raw_termios.c_cc[VTIME] = 0;
     }
-    if (tcsetattr(fd, TCSAFLUSH, &raw_termios) != 0) { PyErr_SetFromErrno(PyExc_OSError); return false; }
+    if (tcsetattr(fd, optional_actions, &raw_termios) != 0) { PyErr_SetFromErrno(PyExc_OSError); return false; }
     return true;
 }
 
 static PyObject*
 open_tty(PyObject *self UNUSED, PyObject *args) {
-    int read_with_timeout = 0;
-    if (!PyArg_ParseTuple(args, "|p", &read_with_timeout)) return NULL;
+    int read_with_timeout = 0, optional_actions = TCSAFLUSH;
+    if (!PyArg_ParseTuple(args, "|pi", &read_with_timeout, &optional_actions)) return NULL;
     int flags = O_RDWR | O_CLOEXEC | O_NOCTTY;
     if (!read_with_timeout) flags |= O_NONBLOCK;
     static char ctty[L_ctermid+1];
-    int fd = open(ctermid(ctty), flags);
+    int fd = safe_open(ctermid(ctty), flags, 0);
     if (fd == -1) { PyErr_Format(PyExc_OSError, "Failed to open controlling terminal: %s (identified with ctermid()) with error: %s", ctty, strerror(errno)); return NULL; }
     struct termios *termios_p = calloc(1, sizeof(struct termios));
     if (!termios_p) return PyErr_NoMemory();
     if (tcgetattr(fd, termios_p) != 0) { free(termios_p); PyErr_SetFromErrno(PyExc_OSError); return NULL; }
-    if (!put_tty_in_raw_mode(fd, termios_p, read_with_timeout != 0)) { free(termios_p); return NULL; }
+    if (!put_tty_in_raw_mode(fd, termios_p, read_with_timeout != 0, optional_actions)) { free(termios_p); return NULL; }
     return Py_BuildValue("iN", fd, PyLong_FromVoidPtr(termios_p));
 }
 
 #define TTY_ARGS \
-    PyObject *tp; int fd; \
-    if (!PyArg_ParseTuple(args, "iO!", &fd, &PyLong_Type, &tp)) return NULL; \
+    PyObject *tp; int fd; int optional_actions = TCSAFLUSH; \
+    if (!PyArg_ParseTuple(args, "iO!|i", &fd, &PyLong_Type, &tp, &optional_actions)) return NULL; \
     struct termios *termios_p = PyLong_AsVoidPtr(tp);
 
 static PyObject*
 normal_tty(PyObject *self UNUSED, PyObject *args) {
     TTY_ARGS
-    if (tcsetattr(fd, TCSAFLUSH, termios_p) != 0) { PyErr_SetFromErrno(PyExc_OSError); return NULL; }
+    if (tcsetattr(fd, optional_actions, termios_p) != 0) { PyErr_SetFromErrno(PyExc_OSError); return NULL; }
     Py_RETURN_NONE;
 }
 
 static PyObject*
 raw_tty(PyObject *self UNUSED, PyObject *args) {
     TTY_ARGS
-    if (!put_tty_in_raw_mode(fd, termios_p, false)) return NULL;
+    if (!put_tty_in_raw_mode(fd, termios_p, false, optional_actions)) return NULL;
     Py_RETURN_NONE;
 }
 
@@ -140,15 +143,23 @@ raw_tty(PyObject *self UNUSED, PyObject *args) {
 static PyObject*
 close_tty(PyObject *self UNUSED, PyObject *args) {
     TTY_ARGS
-    tcsetattr(fd, TCSAFLUSH, termios_p);  // deliberately ignore failure
+    tcsetattr(fd, optional_actions, termios_p);  // deliberately ignore failure
     free(termios_p);
-    safe_close(fd);
+    safe_close(fd, __FILE__, __LINE__);
     Py_RETURN_NONE;
 }
 
 #undef TTY_ARGS
 
+static PyObject*
+wcwidth_wrap(PyObject UNUSED *self, PyObject *chr) {
+    return PyLong_FromLong(wcwidth_std(PyLong_AsLong(chr)));
+}
+
+
 static PyMethodDef module_methods[] = {
+    {"wcwidth", (PyCFunction)wcwidth_wrap, METH_O, ""},
+    {"wcswidth", (PyCFunction)wcswidth_std, METH_O, ""},
     {"open_tty", open_tty, METH_VARARGS, ""},
     {"normal_tty", normal_tty, METH_VARARGS, ""},
     {"raw_tty", raw_tty, METH_VARARGS, ""},
@@ -182,6 +193,7 @@ static struct PyModuleDef module = {
 extern int init_LineBuf(PyObject *);
 extern int init_HistoryBuf(PyObject *);
 extern int init_Cursor(PyObject *);
+extern int init_DiskCache(PyObject *);
 extern bool init_child_monitor(PyObject *);
 extern int init_Line(PyObject *);
 extern int init_ColorProfile(PyObject *);
@@ -205,6 +217,7 @@ extern bool init_cocoa(PyObject *module);
 extern bool init_macos_process_info(PyObject *module);
 #else
 extern bool init_freetype_library(PyObject*);
+extern bool init_freetype_render_ui_text(PyObject*);
 #endif
 
 
@@ -214,61 +227,65 @@ PyInit_fast_data_types(void) {
 
     m = PyModule_Create(&module);
     if (m == NULL) return NULL;
+    if (Py_AtExit(run_at_exit_cleanup_functions) != 0) {
+        PyErr_SetString(PyExc_RuntimeError, "Failed to register the atexit cleanup handler");
+        return NULL;
+    }
     init_monotonic();
 
-    if (m != NULL) {
-        if (!init_logging(m)) return NULL;
-        if (!init_LineBuf(m)) return NULL;
-        if (!init_HistoryBuf(m)) return NULL;
-        if (!init_Line(m)) return NULL;
-        if (!init_Cursor(m)) return NULL;
-        if (!init_child_monitor(m)) return NULL;
-        if (!init_ColorProfile(m)) return NULL;
-        if (!init_Screen(m)) return NULL;
-        if (!init_glfw(m)) return NULL;
-        if (!init_child(m)) return NULL;
-        if (!init_state(m)) return NULL;
-        if (!init_keys(m)) return NULL;
-        if (!init_graphics(m)) return NULL;
-        if (!init_shaders(m)) return NULL;
-        if (!init_mouse(m)) return NULL;
-        if (!init_kittens(m)) return NULL;
-        if (!init_png_reader(m)) return NULL;
+    if (!init_logging(m)) return NULL;
+    if (!init_LineBuf(m)) return NULL;
+    if (!init_HistoryBuf(m)) return NULL;
+    if (!init_Line(m)) return NULL;
+    if (!init_Cursor(m)) return NULL;
+    if (!init_DiskCache(m)) return NULL;
+    if (!init_child_monitor(m)) return NULL;
+    if (!init_ColorProfile(m)) return NULL;
+    if (!init_Screen(m)) return NULL;
+    if (!init_glfw(m)) return NULL;
+    if (!init_child(m)) return NULL;
+    if (!init_state(m)) return NULL;
+    if (!init_keys(m)) return NULL;
+    if (!init_graphics(m)) return NULL;
+    if (!init_shaders(m)) return NULL;
+    if (!init_mouse(m)) return NULL;
+    if (!init_kittens(m)) return NULL;
+    if (!init_png_reader(m)) return NULL;
 #ifdef __APPLE__
-        if (!init_macos_process_info(m)) return NULL;
-        if (!init_CoreText(m)) return NULL;
-        if (!init_cocoa(m)) return NULL;
+    if (!init_macos_process_info(m)) return NULL;
+    if (!init_CoreText(m)) return NULL;
+    if (!init_cocoa(m)) return NULL;
 #else
-        if (!init_freetype_library(m)) return NULL;
-        if (!init_fontconfig_library(m)) return NULL;
-        if (!init_desktop(m)) return NULL;
+    if (!init_freetype_library(m)) return NULL;
+    if (!init_fontconfig_library(m)) return NULL;
+    if (!init_desktop(m)) return NULL;
+    if (!init_freetype_render_ui_text(m)) return NULL;
 #endif
-        if (!init_fonts(m)) return NULL;
+    if (!init_fonts(m)) return NULL;
 
-        PyModule_AddIntConstant(m, "BOLD", BOLD_SHIFT);
-        PyModule_AddIntConstant(m, "ITALIC", ITALIC_SHIFT);
-        PyModule_AddIntConstant(m, "REVERSE", REVERSE_SHIFT);
-        PyModule_AddIntConstant(m, "STRIKETHROUGH", STRIKE_SHIFT);
-        PyModule_AddIntConstant(m, "DIM", DIM_SHIFT);
-        PyModule_AddIntConstant(m, "DECORATION", DECORATION_SHIFT);
-        PyModule_AddIntConstant(m, "MARK", MARK_SHIFT);
-        PyModule_AddIntConstant(m, "MARK_MASK", MARK_MASK);
-        PyModule_AddStringMacro(m, ERROR_PREFIX);
+    PyModule_AddIntConstant(m, "BOLD", BOLD_SHIFT);
+    PyModule_AddIntConstant(m, "ITALIC", ITALIC_SHIFT);
+    PyModule_AddIntConstant(m, "REVERSE", REVERSE_SHIFT);
+    PyModule_AddIntConstant(m, "STRIKETHROUGH", STRIKE_SHIFT);
+    PyModule_AddIntConstant(m, "DIM", DIM_SHIFT);
+    PyModule_AddIntConstant(m, "DECORATION", DECORATION_SHIFT);
+    PyModule_AddIntConstant(m, "MARK", MARK_SHIFT);
+    PyModule_AddIntConstant(m, "MARK_MASK", MARK_MASK);
+    PyModule_AddStringMacro(m, ERROR_PREFIX);
 #ifdef KITTY_VCS_REV
-        PyModule_AddStringMacro(m, KITTY_VCS_REV);
+    PyModule_AddStringMacro(m, KITTY_VCS_REV);
 #endif
-        PyModule_AddIntMacro(m, CURSOR_BLOCK);
-        PyModule_AddIntMacro(m, CURSOR_BEAM);
-        PyModule_AddIntMacro(m, CURSOR_UNDERLINE);
-        PyModule_AddIntMacro(m, DECAWM);
-        PyModule_AddIntMacro(m, DECCOLM);
-        PyModule_AddIntMacro(m, DECOM);
-        PyModule_AddIntMacro(m, IRM);
-        PyModule_AddIntMacro(m, CSI);
-        PyModule_AddIntMacro(m, DCS);
-        PyModule_AddIntMacro(m, APC);
-        PyModule_AddIntMacro(m, OSC);
-    }
+    PyModule_AddIntMacro(m, CURSOR_BLOCK);
+    PyModule_AddIntMacro(m, CURSOR_BEAM);
+    PyModule_AddIntMacro(m, CURSOR_UNDERLINE);
+    PyModule_AddIntMacro(m, DECAWM);
+    PyModule_AddIntMacro(m, DECCOLM);
+    PyModule_AddIntMacro(m, DECOM);
+    PyModule_AddIntMacro(m, IRM);
+    PyModule_AddIntMacro(m, CSI);
+    PyModule_AddIntMacro(m, DCS);
+    PyModule_AddIntMacro(m, APC);
+    PyModule_AddIntMacro(m, OSC);
 
     return m;
 }
