@@ -6,10 +6,12 @@
  */
 
 #include "state.h"
+#include "cleanup.h"
 #include "lineops.h"
 #include "fonts.h"
 #include <fontconfig/fontconfig.h>
 #include "emoji.h"
+#include "freetype_render_ui_text.h"
 #ifndef FC_COLOR
 #define FC_COLOR "color"
 #endif
@@ -27,17 +29,33 @@ pyspacing(int val) {
 
 static inline PyObject*
 pattern_as_dict(FcPattern *pat) {
-    PyObject *ans = PyDict_New();
+    PyObject *ans = PyDict_New(), *p = NULL, *list = NULL;
     if (ans == NULL) return NULL;
+
 #define PS(x) PyUnicode_Decode((const char*)x, strlen((const char*)x), "UTF-8", "replace")
+
 #define G(type, get, which, conv, name) { \
-    type out; PyObject *p; \
+    type out; \
     if (get(pat, which, 0, &out) == FcResultMatch) { \
-        p = conv(out); if (p == NULL) { Py_CLEAR(ans); return NULL; } \
-        if (PyDict_SetItemString(ans, #name, p) != 0) { Py_CLEAR(p); Py_CLEAR(ans); return NULL; } \
+        p = conv(out); if (p == NULL) goto exit; \
+        if (PyDict_SetItemString(ans, #name, p) != 0) goto exit; \
         Py_CLEAR(p); \
     }}
+
+#define L(type, get, which, conv, name) { \
+    type out; int n = 0; \
+    list = PyList_New(0); \
+    if (!list) goto exit; \
+    while (get(pat, which, n++, &out) == FcResultMatch) { \
+        p = conv(out); if (p == NULL) goto exit; \
+        if (PyList_Append(list, p) != 0) goto exit; \
+        Py_CLEAR(p); \
+    } \
+    if (PyDict_SetItemString(ans, #name, list) != 0) goto exit; \
+    Py_CLEAR(list); \
+}
 #define S(which, key) G(FcChar8*, FcPatternGetString, which, PS, key)
+#define LS(which, key) L(FcChar8*, FcPatternGetString, which, PS, key)
 #define I(which, key) G(int, FcPatternGetInteger, which, PyLong_FromLong, key)
 #define B(which, key) G(int, FcPatternGetBool, which, pybool, key)
 #define E(which, key, conv) G(int, FcPatternGetInteger, which, conv, key)
@@ -46,7 +64,9 @@ pattern_as_dict(FcPattern *pat) {
     S(FC_STYLE, style);
     S(FC_FULLNAME, full_name);
     S(FC_POSTSCRIPT_NAME, postscript_name);
+    LS(FC_FONT_FEATURES, fontfeatures);
     I(FC_WEIGHT, weight);
+    I(FC_WIDTH, width)
     I(FC_SLANT, slant);
     I(FC_HINT_STYLE, hint_style);
     I(FC_INDEX, index);
@@ -57,6 +77,10 @@ pattern_as_dict(FcPattern *pat) {
     B(FC_OUTLINE, outline);
     B(FC_COLOR, color);
     E(FC_SPACING, spacing, pyspacing);
+exit:
+    if (PyErr_Occurred()) Py_CLEAR(ans);
+    Py_CLEAR(p);
+    Py_CLEAR(list);
 
     return ans;
 #undef PS
@@ -65,6 +89,8 @@ pattern_as_dict(FcPattern *pat) {
 #undef B
 #undef E
 #undef G
+#undef L
+#undef LS
 }
 
 static inline PyObject*
@@ -124,7 +150,7 @@ end:
     return ans;
 }
 
-static Py_UCS4 char_buf[1024];
+static char_type char_buf[1024];
 
 static inline void
 add_charset(FcPattern *pat, size_t num) {
@@ -144,6 +170,48 @@ end:
     if (charset != NULL) FcCharSetDestroy(charset);
 }
 
+static inline bool
+_native_fc_match(FcPattern *pat, FontConfigFace *ans) {
+    bool ok = false;
+    FcPattern *match = NULL;
+    FcResult result;
+    FcConfigSubstitute(NULL, pat, FcMatchPattern);
+    FcDefaultSubstitute(pat);
+    /* printf("fc_match = %s\n", FcNameUnparse(pat)); */
+    match = FcFontMatch(NULL, pat, &result);
+    if (match == NULL) { PyErr_SetString(PyExc_KeyError, "FcFontMatch() failed"); goto end; }
+    FcChar8 *out;
+#define g(func, prop, output) if (func(match, prop, 0, &output) != FcResultMatch) { PyErr_SetString(PyExc_ValueError, "No " #prop " found in fontconfig match result"); goto end; }
+    g(FcPatternGetString, FC_FILE, out);
+    g(FcPatternGetInteger, FC_INDEX, ans->index);
+    g(FcPatternGetInteger, FC_HINT_STYLE, ans->hintstyle);
+    g(FcPatternGetBool, FC_HINTING, ans->hinting);
+#undef g
+    ans->path = strdup((char*)out);
+    if (!ans->path) { PyErr_NoMemory(); goto end; }
+    ok = true;
+end:
+    if (match != NULL) FcPatternDestroy(match);
+    return ok;
+}
+
+
+bool
+information_for_font_family(const char *family, bool bold, bool italic, FontConfigFace *ans) {
+    memset(ans, 0, sizeof(FontConfigFace));
+    FcPattern *pat = FcPatternCreate();
+    bool ok = false;
+    if (pat == NULL) { PyErr_NoMemory(); return ok; }
+    if (family && strlen(family) > 0) AP(FcPatternAddString, FC_FAMILY, (const FcChar8*)family, "family");
+    if (bold) { AP(FcPatternAddInteger, FC_WEIGHT, FC_WEIGHT_BOLD, "weight"); }
+    if (italic) { AP(FcPatternAddInteger, FC_SLANT, FC_SLANT_ITALIC, "slant"); }
+    ok = _native_fc_match(pat, ans);
+end:
+    if (pat != NULL) FcPatternDestroy(pat);
+    return ok;
+}
+
+
 static PyObject*
 fc_match(PyObject UNUSED *self, PyObject *args) {
     char *family = NULL;
@@ -178,6 +246,27 @@ end:
     return ans;
 }
 
+static PyObject*
+fc_match_postscript_name(PyObject UNUSED *self, PyObject *args) {
+    const char *postscript_name = NULL;
+    FcPattern *pat = NULL;
+    PyObject *ans = NULL;
+
+    if (!PyArg_ParseTuple(args, "s", &postscript_name)) return NULL;
+    if (!postscript_name || !postscript_name[0]) { PyErr_SetString(PyExc_KeyError, "postscript_name must not be empty"); return NULL; }
+
+    pat = FcPatternCreate();
+    if (pat == NULL) return PyErr_NoMemory();
+
+    AP(FcPatternAddString, FC_POSTSCRIPT_NAME, (const FcChar8*)postscript_name, "postscript_name");
+
+    ans = _fc_match(pat);
+
+end:
+    if (pat != NULL) FcPatternDestroy(pat);
+    return ans;
+}
+
 PyObject*
 specialize_font_descriptor(PyObject *base_descriptor, FONTS_DATA_HANDLE fg) {
     PyObject *p = PyDict_GetItemString(base_descriptor, "path"), *ans = NULL;
@@ -201,6 +290,24 @@ end:
     return ans;
 }
 
+bool
+fallback_font(char_type ch, const char *family, bool bold, bool italic, bool prefer_color, FontConfigFace *ans) {
+    memset(ans, 0, sizeof(FontConfigFace));
+    bool ok = false;
+    FcPattern *pat = FcPatternCreate();
+    if (pat == NULL) { PyErr_NoMemory(); return ok; }
+    if (family) AP(FcPatternAddString, FC_FAMILY, (const FcChar8*)family, "family");
+    if (bold) { AP(FcPatternAddInteger, FC_WEIGHT, FC_WEIGHT_BOLD, "weight"); }
+    if (italic) { AP(FcPatternAddInteger, FC_SLANT, FC_SLANT_ITALIC, "slant"); }
+    if (prefer_color) { AP(FcPatternAddBool, FC_COLOR, true, "color"); }
+    char_buf[0] = ch;
+    add_charset(pat, 1);
+    ok = _native_fc_match(pat, ans);
+end:
+    if (pat != NULL) FcPatternDestroy(pat);
+    return ok;
+}
+
 PyObject*
 create_fallback_face(PyObject UNUSED *base_face, CPUCell* cell, bool bold, bool italic, bool emoji_presentation, FONTS_DATA_HANDLE fg) {
     PyObject *ans = NULL;
@@ -223,6 +330,7 @@ end:
 static PyMethodDef module_methods[] = {
     METHODB(fc_list, METH_VARARGS),
     METHODB(fc_match, METH_VARARGS),
+    METHODB(fc_match_postscript_name, METH_VARARGS),
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
@@ -232,10 +340,7 @@ init_fontconfig_library(PyObject *module) {
         PyErr_SetString(PyExc_RuntimeError, "Failed to initialize the fontconfig library");
         return false;
     }
-    if (Py_AtExit(FcFini) != 0) {
-        PyErr_SetString(PyExc_RuntimeError, "Failed to register the fontconfig library at exit handler");
-        return false;
-    }
+    register_at_exit_cleanup_func(FONTCONFIG_CLEANUP_FUNC, FcFini);
     if (PyModule_AddFunctions(module, module_methods) != 0) return false;
     PyModule_AddIntMacro(module, FC_WEIGHT_REGULAR);
     PyModule_AddIntMacro(module, FC_WEIGHT_MEDIUM);
@@ -247,5 +352,7 @@ init_fontconfig_library(PyObject *module) {
     PyModule_AddIntMacro(module, FC_DUAL);
     PyModule_AddIntMacro(module, FC_MONO);
     PyModule_AddIntMacro(module, FC_CHARCELL);
+    PyModule_AddIntMacro(module, FC_WIDTH_NORMAL);
+
     return true;
 }
