@@ -3,17 +3,17 @@
 # License: GPLv3 Copyright: 2019, Kovid Goyal <kovid at kovidgoyal.net>
 
 
-from functools import lru_cache
-from typing import Any, Dict, List, Optional, Sequence, Tuple
+from typing import Any, Dict, List, NamedTuple, Optional, Sequence
 
 from .boss import Boss
 from .child import Child
-from .cli import parse_args
+from .cli import WATCHER_DEFINITION, parse_args
 from .cli_stub import LaunchCLIOptions
 from .constants import resolve_custom_file
-from .fast_data_types import set_clipboard_string
+from .fast_data_types import patch_color_profiles, set_clipboard_string
 from .tabs import Tab
-from .utils import set_primary_selection
+from .types import run_once
+from .utils import find_exe, read_shell_environment, set_primary_selection
 from .window import Watchers, Window
 
 try:
@@ -22,7 +22,12 @@ except ImportError:
     TypedDict = Dict[str, Any]
 
 
-@lru_cache(maxsize=2)
+class LaunchSpec(NamedTuple):
+    opts: LaunchCLIOptions
+    args: List[str]
+
+
+@run_once
 def options_spec() -> str:
     return '''
 --window-title --title
@@ -42,7 +47,7 @@ choices=window,tab,os-window,overlay,background,clipboard,primary
 Where to launch the child process, in a new kitty window in the current tab,
 a new tab, or a new OS window or an overlay over the current window.
 Note that if the current window already has an overlay, then it will
-open a new window. The value of none means the process will be
+open a new window. The value of background means the process will be
 run in the background. The values clipboard and primary are meant
 to work with :option:`launch --stdin-source` to copy data to the system
 clipboard or primary selection.
@@ -81,13 +86,17 @@ currently active window.
 --copy-env
 type=bool-set
 Copy the environment variables from the currently active window into the
-newly launched child process.
+newly launched child process. Note that most shells only set environment
+variables for child processes, so this will only copy the environment
+variables that the shell process itself has not the environment variables
+child processes inside the shell see. To copy that enviroment, use the
+kitty remote control feature with :code:`kitty @launch --copy-env`.
 
 
 --location
 type=choices
-default=last
-choices=first,after,before,neighbor,last,vsplit,hsplit
+default=default
+choices=first,after,before,neighbor,last,vsplit,hsplit,default
 Where to place the newly created window when it is added to a tab which
 already has existing windows in it. :code:`after` and :code:`before` place the new
 window before or after the active window. :code:`neighbor` is a synonym for :code:`after`.
@@ -95,7 +104,8 @@ Also applies to creating a new tab, where the value of :code:`after`
 will cause the new tab to be placed next to the current tab instead of at the end.
 The values of :code:`vsplit` and :code:`hsplit` are only used by the :code:`splits`
 layout and control if the new window is placed in a vertical or horizontal split
-with the currently active window.
+with the currently active window. The default is to place the window in a
+layout dependent manner, typically, after the currently active window.
 
 
 --allow-remote-control
@@ -111,11 +121,11 @@ computers (for example, over ssh) or as other users.
 type=choices
 default=none
 choices=none,@selection,@screen,@screen_scrollback,@alternate,@alternate_scrollback
-Pass the screen contents as :code:`STDIN` to the child process. @selection is
-the currently selected text. @screen is the contents of the currently active
-window. @screen_scrollback is the same as @screen, but includes the scrollback
-buffer as well. @alternate is the secondary screen of the current active
-window. For example if you run a full screen terminal application, the
+Pass the screen contents as :code:`STDIN` to the child process. :code:`@selection` is
+the currently selected text. :code:`@screen` is the contents of the currently active
+window. :code:`@screen_scrollback` is the same as :code:`@screen`, but includes the
+scrollback buffer as well. :code:`@alternate` is the secondary screen of the current
+active window. For example if you run a full screen terminal application, the
 secondary screen will be the screen you return to when quitting the
 application.
 
@@ -151,22 +161,23 @@ Set the WM_NAME property on X11 for the newly created OS Window when using
 :option:`launch --type`=os-window. Defaults to :option:`launch --os-window-class`.
 
 
---watcher -w
+--color
 type=list
-Path to a python file. Appropriately named functions in this file will be called
-for various events, such as when the window is resized or closed. See the section
-on watchers in the launch command documentation :doc:`launch`. Relative paths are
-resolved relative to the kitty config directory.
-'''
+Change colors in the newly launched window. You can either specify a path to a .conf
+file with the same syntax as kitty.conf to read the colors from, or specify them
+individually, for example: ``--color background=white`` ``--color foreground=red``
 
 
-def parse_launch_args(args: Optional[Sequence[str]] = None) -> Tuple[LaunchCLIOptions, List[str]]:
+''' + WATCHER_DEFINITION
+
+
+def parse_launch_args(args: Optional[Sequence[str]] = None) -> LaunchSpec:
     args = list(args or ())
     try:
         opts, args = parse_args(result_class=LaunchCLIOptions, args=args, ospec=options_spec)
     except SystemExit as e:
         raise ValueError from e
-    return opts, args
+    return LaunchSpec(opts, args)
 
 
 def get_env(opts: LaunchCLIOptions, active_child: Child) -> Dict[str, str]:
@@ -201,20 +212,23 @@ def tab_for_window(boss: Boss, opts: LaunchCLIOptions, target_tab: Optional[Tab]
     return tab
 
 
-def load_watch_modules(opts: LaunchCLIOptions) -> Optional[Watchers]:
-    if not opts.watcher:
+def load_watch_modules(watchers: Sequence[str]) -> Optional[Watchers]:
+    if not watchers:
         return None
     import runpy
     ans = Watchers()
-    for path in opts.watcher:
+    for path in watchers:
         path = resolve_custom_file(path)
-        m = runpy.run_path(path, run_name='__kitty_watcher__')
+        m = runpy.run_path(path, run_name='__kitty_watcher__')  # type: ignore
         w = m.get('on_close')
         if callable(w):
             ans.on_close.append(w)
         w = m.get('on_resize')
         if callable(w):
             ans.on_resize.append(w)
+        w = m.get('on_focus_change')
+        if callable(w):
+            ans.on_focus_change.append(w)
     return ans
 
 
@@ -232,7 +246,20 @@ class LaunchKwds(TypedDict):
     stdin: Optional[bytes]
 
 
-def launch(boss: Boss, opts: LaunchCLIOptions, args: List[str], target_tab: Optional[Tab] = None) -> Optional[Window]:
+def apply_colors(window: Window, spec: Sequence[str]) -> None:
+    from kitty.rc.set_colors import parse_colors
+    colors, cursor_text_color = parse_colors(spec)
+    profiles = window.screen.color_profile,
+    patch_color_profiles(colors, cursor_text_color, profiles, True)
+
+
+def launch(
+    boss: Boss,
+    opts: LaunchCLIOptions,
+    args: List[str],
+    target_tab: Optional[Tab] = None,
+    force_target_tab: bool = False
+) -> Optional[Window]:
     active = boss.active_window_for_cwd
     active_child = getattr(active, 'child', None)
     env = get_env(opts, active_child)
@@ -254,7 +281,7 @@ def launch(boss: Boss, opts: LaunchCLIOptions, args: List[str], target_tab: Opti
                 kw['cwd_from'] = active_child.pid_for_cwd
         else:
             kw['cwd'] = opts.cwd
-    if opts.location != 'last':
+    if opts.location != 'default':
         kw['location'] = opts.location
     if opts.copy_colors and active:
         kw['copy_colors_from'] = active
@@ -272,11 +299,19 @@ def launch(boss: Boss, opts: LaunchCLIOptions, args: List[str], target_tab: Opti
                 elif x == '@active-kitty-window-id':
                     x = str(active.id)
             final_cmd.append(x)
+        exe = find_exe(final_cmd[0])
+        if not exe:
+            env = read_shell_environment(boss.opts)
+            if 'PATH' in env:
+                import shutil
+                exe = shutil.which(final_cmd[0], path=env['PATH'])
+        if exe:
+            final_cmd[0] = exe
         kw['cmd'] = final_cmd
-    if opts.type == 'overlay' and active and not active.overlay_window_id:
+    if opts.type == 'overlay' and active:
         kw['overlay_for'] = active.id
-    if opts.stdin_source and opts.stdin_source != 'none':
-        q = opts.stdin_source
+    if opts.stdin_source != 'none':
+        q = str(opts.stdin_source)
         if opts.stdin_add_formatting:
             if q in ('@screen', '@screen_scrollback', '@alternate', '@alternate_scrollback'):
                 q = '@ansi_' + q[1:]
@@ -301,10 +336,15 @@ def launch(boss: Boss, opts: LaunchCLIOptions, args: List[str], target_tab: Opti
             else:
                 set_primary_selection(stdin)
     else:
-        tab = tab_for_window(boss, opts, target_tab)
+        if force_target_tab:
+            tab = target_tab
+        else:
+            tab = tab_for_window(boss, opts, target_tab)
         if tab is not None:
-            watchers = load_watch_modules(opts)
+            watchers = load_watch_modules(opts.watcher)
             new_window: Window = tab.new_window(env=env or None, watchers=watchers or None, **kw)
+            if opts.color:
+                apply_colors(new_window, opts.color)
             if opts.keep_focus and active:
                 boss.set_active_window(active, switch_os_window_if_needed=True)
             return new_window
