@@ -195,6 +195,14 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
 
 #define HANDLER(name) static inline void name(Window UNUSED *w, int UNUSED button, int UNUSED modifiers, unsigned int UNUSED window_idx)
 
+static inline void
+set_mouse_cursor_when_dragging(void) {
+    if (mouse_cursor_shape != OPT(pointer_shape_when_dragging)) {
+        mouse_cursor_shape = OPT(pointer_shape_when_dragging);
+        set_mouse_cursor(mouse_cursor_shape);
+    }
+}
+
 static inline void
 update_drag(bool from_button, Window *w, bool is_release, int modifiers) {
     Screen *screen = w->render_data.screen;
@@ -202,16 +210,18 @@ update_drag(bool from_button, Window *w, bool is_release, int modifiers) {
         if (is_release) {
             global_state.active_drag_in_window = 0;
             w->last_drag_scroll_at = 0;
-            if (screen->selection.in_progress)
+            if (screen->selections.in_progress) {
                 screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, true, false);
+            }
         }
         else {
             global_state.active_drag_in_window = w->id;
             screen_start_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, modifiers == (int)OPT(rectangle_select_modifiers) || modifiers == ((int)OPT(rectangle_select_modifiers) | (int)OPT(terminal_select_modifiers)), EXTEND_CELL);
         }
-    } else if (screen->selection.in_progress) {
+    } else if (screen->selections.in_progress) {
         screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, false, false);
     }
+    set_mouse_cursor_when_dragging();
 }
 
 static inline bool
@@ -249,6 +259,7 @@ extend_selection(Window *w, bool ended) {
     if (screen_has_selection(screen)) {
         screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, ended, false);
     }
+    set_mouse_cursor_when_dragging();
 }
 
 static inline void
@@ -264,7 +275,7 @@ extend_url(Screen *screen, Line *line, index_type *x, index_type *y, char_type s
         // we deliberately allow non-continued lines as some programs, like
         // mutt split URLs with newlines at line boundaries
         index_type new_x = line_url_end_at(line, 0, false, sentinel, next_line_starts_with_url_chars);
-        if (!new_x) break;
+        if (!new_x && !line_startswith_url_chars(line)) break;
         *y += 1; *x = new_x;
     }
 }
@@ -294,7 +305,7 @@ get_url_sentinel(Line *line, index_type url_start) {
 
 static inline void
 set_mouse_cursor_for_screen(Screen *screen) {
-    mouse_cursor_shape = screen->modes.mouse_tracking_mode == NO_TRACKING ? BEAM : OPT(pointer_shape_when_grabbed);
+    mouse_cursor_shape = screen->modes.mouse_tracking_mode == NO_TRACKING ? OPT(default_pointer_shape): OPT(pointer_shape_when_grabbed);
 }
 
 static inline void
@@ -302,6 +313,11 @@ detect_url(Screen *screen, unsigned int x, unsigned int y) {
     bool has_url = false;
     index_type url_start, url_end = 0;
     Line *line = screen_visual_line(screen, y);
+    if (line->cpu_cells[x].hyperlink_id) {
+        mouse_cursor_shape = HAND;
+        screen_mark_hyperlink(screen, x, y);
+        return;
+    }
     char_type sentinel;
     if (line) {
         url_start = line_url_start_at(line, x);
@@ -331,7 +347,7 @@ detect_url(Screen *screen, unsigned int x, unsigned int y) {
 static inline void
 handle_mouse_movement_in_kitty(Window *w, int button, bool mouse_cell_changed) {
     Screen *screen = w->render_data.screen;
-    if (screen->selection.in_progress && (button == GLFW_MOUSE_BUTTON_LEFT || button == GLFW_MOUSE_BUTTON_RIGHT)) {
+    if (screen->selections.in_progress && (button == GLFW_MOUSE_BUTTON_LEFT || button == GLFW_MOUSE_BUTTON_RIGHT)) {
         monotonic_t now = monotonic();
         if ((now - w->last_drag_scroll_at) >= ms_to_monotonic_t(20ll) || mouse_cell_changed) {
             update_drag(false, w, false, 0);
@@ -342,6 +358,7 @@ handle_mouse_movement_in_kitty(Window *w, int button, bool mouse_cell_changed) {
 }
 
 HANDLER(handle_move_event) {
+    modifiers &= ~GLFW_LOCK_MASK;
     unsigned int x = 0, y = 0;
     if (OPT(focus_follows_mouse)) {
         Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
@@ -352,7 +369,7 @@ HANDLER(handle_move_event) {
     bool in_left_half_of_cell = false;
     if (!cell_for_pos(w, &x, &y, &in_left_half_of_cell, global_state.callback_os_window)) return;
     Screen *screen = w->render_data.screen;
-    detect_url(screen, x, y);
+    if(OPT(detect_urls)) detect_url(screen, x, y);
     bool mouse_cell_changed = x != w->mouse_pos.cell_x || y != w->mouse_pos.cell_y;
     bool cell_half_changed = in_left_half_of_cell != w->mouse_pos.in_left_half_of_cell;
     w->mouse_pos.cell_x = x; w->mouse_pos.cell_y = y;
@@ -391,6 +408,7 @@ multi_click(Window *w, unsigned int count) {
         screen_start_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, false, mode);
         screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, false, true);
     }
+    set_mouse_cursor_when_dragging();
 }
 
 static inline double
@@ -404,6 +422,7 @@ clear_click_queue(Window *w) {
 }
 
 HANDLER(add_click) {
+    modifiers &= ~GLFW_LOCK_MASK;
     ClickQueue *q = &w->click_queue;
     if (q->length == CLICK_QUEUE_SZ) { memmove(q->clicks, q->clicks + 1, sizeof(Click) * (CLICK_QUEUE_SZ - 1)); q->length--; }
     monotonic_t now = monotonic();
@@ -460,6 +479,7 @@ handle_button_event_in_kitty(Window *w, int button, int modifiers, bool is_relea
 }
 
 HANDLER(handle_button_event) {
+    modifiers &= ~GLFW_LOCK_MASK;
     Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
     bool is_release = !global_state.callback_os_window->mouse_button_pressed[button];
     if (window_idx != t->active_window && !is_release) {
@@ -489,6 +509,7 @@ currently_pressed_button(void) {
 }
 
 HANDLER(handle_event) {
+    modifiers &= ~GLFW_LOCK_MASK;
     if (button == -1) {
         button = currently_pressed_button();
         handle_move_event(w, button, modifiers, window_idx);
@@ -499,8 +520,12 @@ HANDLER(handle_event) {
 
 static inline void
 handle_tab_bar_mouse(int button, int UNUSED modifiers) {
+    static monotonic_t last_click_at = 0;
     if (button != GLFW_MOUSE_BUTTON_LEFT || !global_state.callback_os_window->mouse_button_pressed[button]) return;
-    call_boss(activate_tab_at, "Kd", global_state.callback_os_window->id, global_state.callback_os_window->mouse_x);
+    monotonic_t now = monotonic();
+    bool is_double = now - last_click_at <= OPT(click_interval);
+    last_click_at = is_double ? 0 : now;
+    call_boss(activate_tab_at, "KdO", global_state.callback_os_window->id, global_state.callback_os_window->mouse_x, is_double ? Py_True : Py_False);
 }
 
 static inline bool
@@ -685,9 +710,8 @@ scroll_event(double UNUSED xoffset, double yoffset, int flags, int modifiers) {
         s = (int)round(pixels) / (int)global_state.callback_os_window->fonts_data->cell_height;
         screen->pending_scroll_pixels = pixels - s * (int) global_state.callback_os_window->fonts_data->cell_height;
     } else {
-        if (screen->linebuf == screen->main_linebuf || !screen->modes.mouse_tracking_mode) {
-            // Only use wheel_scroll_multiplier if we are scrolling kitty scrollback or in mouse
-            // tracking mode, where the application is responsible for interpreting scroll events
+        if (!screen->modes.mouse_tracking_mode) {
+            // Dont use multiplier if we are sending events to the application
             yoffset *= OPT(wheel_scroll_multiplier);
         } else if (OPT(wheel_scroll_multiplier) < 0) {
             // ensure that changing scroll direction still works, even though
@@ -702,20 +726,17 @@ scroll_event(double UNUSED xoffset, double yoffset, int flags, int modifiers) {
     }
     if (s == 0) return;
     bool upwards = s > 0;
-    if (screen->linebuf == screen->main_linebuf) {
-        screen_history_scroll(screen, abs(s), upwards);
-    } else {
-        if (screen->modes.mouse_tracking_mode) {
-            int sz = encode_mouse_scroll(w, upwards, modifiers);
-            if (sz > 0) {
-                mouse_event_buf[sz] = 0;
-                for (s = abs(s); s > 0; s--) {
-                    write_escape_code_to_child(screen, CSI, mouse_event_buf);
-                }
+    if (screen->modes.mouse_tracking_mode) {
+        int sz = encode_mouse_scroll(w, upwards, modifiers);
+        if (sz > 0) {
+            mouse_event_buf[sz] = 0;
+            for (s = abs(s); s > 0; s--) {
+                write_escape_code_to_child(screen, CSI, mouse_event_buf);
             }
-        } else {
-            fake_scroll(w, abs(s), upwards);
         }
+    } else {
+        if (screen->linebuf == screen->main_linebuf) screen_history_scroll(screen, abs(s), upwards);
+        else fake_scroll(w, abs(s), upwards);
     }
 }
 
